import shlex
import socket

import click.exceptions

from prompt_toolkit import PromptSession, HTML
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.completion import Completer as BaseCompleter, Completion
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.application import run_in_terminal

from icepap import version as sw_version

from .utils import get_axes
from .. import version

# Handle backwards compatibility between Click 7.0 and 8.0
try:
    import click.shell_completion
    HAS_C8 = True
except ImportError:
    import click._bashcomplete
    HAS_C8 = False


def human_host(host):
    if host in {"", "0", "0.0.0.0"}:
        return socket.gethostname()
    return host


def get_addr_ver(icepap):

    if icepap.port != 5000:
        addr = "{}:{}".format(icepap.host, icepap.port)
    else:
        addr = icepap.host

    try:
        ver = icepap.fver
    except Exception:
        ver = 'ERROR: Can not read master version'

    return addr, ver


class Toolbar:
    def __init__(self, icepap):
        self.icepap = icepap
        self.addr, self.ver = get_addr_ver(icepap)

    def __call__(self):
        msg = "icepapctl {} | {} - {}| " \
              "<b>[F5]</b>: State <b>[F6]</b>: Status | " \
              "<b>[Ctrl-D]</b>: Quit".format(sw_version, self.addr,
                                             self.ver)
        return HTML(msg)


class Completer(BaseCompleter):

    def __init__(self, ctx):
        self.ctx = ctx
        self.icepap = ctx.obj["icepap"]

    def get_completions(self, document, complete_event=None):
        # Code analogous to click._bashcomplete.do_complete
        try:
            args = shlex.split(document.text_before_cursor)
        except ValueError:
            # Invalid command, perhaps caused by missing closing quotation.
            return

        cursor_within_command = (
            document.text_before_cursor.rstrip() == document.text_before_cursor
        )

        if args and cursor_within_command:
            # We've entered some text and no space, give completions for the
            # current word.
            incomplete = args.pop()
        else:
            # We've not entered anything, either at all or for the current
            # command, so give all relevant completions for this context.
            incomplete = ""

        # icepapctl specific:
        # assumes a CLI: icepapctl [OPTS] ICEPAP COMMAND [ARGS]...
        if args:
            args.insert(0, self.icepap.host)

        if HAS_C8:
            ctx = click.shell_completion._resolve_context(self.ctx.command,
                                                          {}, "",
                                                          args)
        else:
            ctx = click._bashcomplete.resolve_ctx(self.ctx.command, "", args)
        if ctx is None:
            return

        if args:
            choices = [Completion(
                "--help", -len(incomplete), display_meta="displays help")]
        else:
            choices = [Completion(
                "exit", -len(incomplete), display_meta="Quits application")]

        for param in ctx.command.params:
            if isinstance(param, click.Option):
                for options in (param.opts, param.secondary_opts):
                    for o in options:
                        choices.append(Completion(
                            str(o), -len(incomplete), display_meta=param.help))
            elif isinstance(param, click.Argument):
                if isinstance(param.type, click.Choice):
                    for choice in param.type.choices:
                        choices.append(Completion(
                            str(choice), -len(incomplete)))

        if isinstance(ctx.command, click.MultiCommand):
            for name in ctx.command.list_commands(ctx):
                command = ctx.command.get_command(ctx, name)
                choices.append(
                    Completion(
                        str(name),
                        -len(incomplete),
                        display_meta=command.get_short_help_str(),
                    )
                )

        for item in choices:
            if item.text.startswith(incomplete):
                yield item


def Prompt(context):
    icepap = context.obj["icepap"]

    kb = KeyBindings()

    def state():
        command("state", context)
        print()

    def status():
        command("status", context)
        print()

    @kb.add("f5")
    def _(event):
        run_in_terminal(state)

    @kb.add("f6")
    def _(event):
        run_in_terminal(status)

    @kb.add('c-space')
    def _(event):
        """ Initialize autocompletion, or select the next completion. """
        buff = event.app.current_buffer
        if buff.complete_state:
            buff.complete_next()
        else:
            buff.start_completion(select_first=False)

    return PromptSession(
        completer=Completer(context),
        history=InMemoryHistory(),
        auto_suggest=AutoSuggestFromHistory(),
        bottom_toolbar=Toolbar(icepap),
        key_bindings=kb,
        enable_history_search=True,
        complete_while_typing=False,
        message="> ",
    )


def command(text, context):
    args = shlex.split(text)
    group = context.command
    name = args[0]
    try:
        if name not in group.commands:
            name = 'send'
        else:
            args = args[1:]

        # From click documentation:
        # https://click.palletsprojects.com/en/8.1.x/arguments/
        # Option-Like Arguments:
        # To recognize negative values Click does what any POSIX style
        # command line script does, and that is to accept the string -- as a
        # separator for options and arguments. After the -- marker, all
        # further parameters are accepted as arguments.
        #
        # This is valid for send, mv, mvr
        if name in ['send', 'mv', 'mvr']:
            args.insert(0, '--')
            if name == 'mvr' and '-m' in args:
                args.pop(args.index('-m'))
                args.insert(0, '-m')

        cmd = group.commands[name]
        cmd.main(args, standalone_mode=False,
                 parent=context, default_map=context.params)
    except (click.exceptions.Exit, click.exceptions.Abort):
        pass
    except click.exceptions.ClickException as error:
        error.show()
    except Exception as error:
        print("Unexpected error: {!r}".format(error))


def print_help(text, ctx):
    args = shlex.split(text)
    if len(args) == 1:
        print('The console allows raw command (Icepap commands) \n'
              'without using the "send" command e.g:\n\n'
              '   "1:?vstatus" is equivalent to "send 1:?vstatus"\n\n')
        print('Console Commands:')
        print('  axes        Get/Set default axes used.')
        print('  help <cmd>  Get CLI command help\n\n')
        print('CLI Commands:')
        print(ctx.get_help().split('Commands:')[1])
    else:
        cmd = args[1]
        cmd_help = ctx.command.commands[cmd].get_help(ctx)
        for line in cmd_help.split('\n'):
            if line.startswith('Usage:'):
                continue
            if line.startswith('Options:'):
                break
            print(line)


def step(prompt, context):
    while True:
        text = prompt.prompt('> ')
        # empty text would create a sub-repl.
        # Avoid it by returning to the prompt
        if not text:
            continue
        elif text == "exit":
            raise EOFError
        elif text.startswith("help"):
            print_help(text, context)
            return
        elif text.startswith('axes'):
            new_axes = text.split('axes')[1]
            if not new_axes:
                print(context.obj['axes_str'])
                return
            else:
                context.obj['axes_str'] = new_axes
                context.obj['axes'] = get_axes(context.obj['icepap'],
                                               new_axes)
                print('New axes: {}'.format(new_axes))
                return

        return command(text, context)


def run(context):
    prompt = Prompt(context)
    icepap = context.obj["icepap"]
    print('Icepap Console Application {}'.format(version))
    addr, ver = get_addr_ver(icepap)
    if 'ERROR' in ver:
        ver = '\033[93m{}\033[0m'.format(ver)
    print('Connected to: {} - {}'.format(addr, ver))
    print('Type "help" for more information.')
    while True:
        try:
            step(prompt, context)
        except EOFError:
            # Ctrl-D
            break



# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------
from functools import wraps

__all__ = ['ver122', 'ver1225', 'ver317', 'FirmwareVersion',
           'SUPPORTED_VERSIONS']

# TODO: Only versions supported @ ALBA Synchrotron
ver122 = {'SYSTEM': {'VER': 1.22,
                     'CONTROLLER': {'VER': 1.22,
                                    'DSP': 2.84,
                                    'FPGA': 0.03,
                                    'MCPU0': 0.2,
                                    'MCPU1': 0.2,
                                    'MCPU2': 1.125,
                                    # 'PCB': 0.07
                                    },
                     'DRIVER': {'VER': 1.22
                                },
                     },
          }

ver1225 = {'SYSTEM': {'VER': 1.225,
                      'CONTROLLER': {'VER': 1.225,
                                     'DSP': 2.85,
                                     'FPGA': 0.03,
                                     'MCPU0': 0.23,
                                     'MCPU1': 0.23,
                                     'MCPU2': 1.125,
                                     # 'PCB': 0.07
                                     },
                      'DRIVER': {'VER': 1.225
                                 },
                      },
           }

ver317 = {'SYSTEM': {'VER': 3.17,
                     'CONTROLLER': {'VER': 3.17,
                                    'DSP': 3.67,
                                    'FPGA': 1.0,
                                    'MCPU0': 1.19,
                                    'MCPU1': 1.19,
                                    'MCPU2': 1.125,
                                    # 'PCB': 1.0
                                    },
                     'DRIVER': {'VER': 3.17
                                },
                     },
          }

# TODO: mismatch between VER INFO and VER SAVED for 1.22 version.
SUPPORTED_VERSIONS = {'1.22': ver122,
                      '1.225': ver1225,
                      '3.17': ver317
                      }


def key_error(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except KeyError:
            return None
    return wrapped


# TODO: improve class to restrict info for component
class FirmwareVersion(dict):
    """
    Class to manage the different version numbers for the different components
    of the IcePAP system. A negative value means an error on the data.
    """
    def __init__(self, data, is_axis=False):
        super(FirmwareVersion, self).__init__()
        self.is_axis = is_axis
        for line in data:
            v = line.split(':', 2)
            length = len(line.split(line.lstrip())[0])
            # print 'length = %s' % l
            component = v[0].strip()
            try:
                value = float(v[1].strip())
            except Exception:
                value = -1.0
            # print component, value
            # manage date
            when = ''
            if len(v) > 2:
                when = v[2].strip()
            if length == 0:  # and component not in self.keys():
                self[component] = dict()
                self[component]['VER'] = (value, when)
                component0 = component
            elif length == 3:  # and component not in self['SYSTEM'].keys():
                self[component0][component] = dict()
                self[component0][component]['VER'] = (value, when)
                component1 = component
            elif length == 6:
                self[component0][component1][component] = (value, when)

    def __repr__(self):

        msg = '{0:<15s}:{1:>5s}\n'.format('SYSTEM', str(self.system))
        level = '   '
        sublevel = level * 2
        if not self.is_axis:
            # Not use field name does not work on python 2.6
            msg += '{0:}{1:<12s}:{2:>5s}\n'.format(level, 'CONTROLLER',
                                                   str(self.ctrl))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'DSP',
                                                  str(self.ctrl_dsp))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'FPGA',
                                                  str(self.ctrl_fpga))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'PCB',
                                                  str(self.ctrl_pcb))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'MCPU0',
                                                  str(self.ctrl_mcpu0))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'MCPU1',
                                                  str(self.ctrl_mcpu1))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'MCPU2',
                                                  str(self.ctrl_mcpu2))
        msg += '{0:}{1:<12s}:{2:>5s}\n'.format(level, 'DRIVER',
                                               str(self.driver))
        if self.is_axis:
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'DSP',
                                                  str(self.driver_dsp))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'FPGA',
                                                  str(self.driver_fpga))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'PCB',
                                                  str(self.driver_pcb))
            msg += '{0:}{1:<9s}:{2:>5s}\n'.format(sublevel, 'IO',
                                                  str(self.driver_io))
        return msg

    def is_supported(self):
        """
        Returns is the system configuration is one of the supported versions.

        :return: bool
        """
        system = self._is_valid_system()
        ctrl = self._is_valid_ctrl()
        driver = self._is_valid_driver()
        return system and ctrl and driver

    def _is_valid_system(self):
        # print('supported system', str(self['SYSTEM']['VER'][0]))
        return str(self['SYSTEM']['VER'][0]) in list(SUPPORTED_VERSIONS.keys())

    @key_error
    def _is_valid_ctrl(self):
        if self._is_valid_system():
            _sys = str(self.system[0])
            _ctrl_ver = SUPPORTED_VERSIONS[_sys]['SYSTEM']['CONTROLLER']['VER']
            if self['SYSTEM']['CONTROLLER']['VER'][0] == _ctrl_ver:
                d = self['SYSTEM']['CONTROLLER']
                a = SUPPORTED_VERSIONS[_sys]['SYSTEM']['CONTROLLER']
                # Does not work on python 2.6
                # _d = {x: d[x][0] for x in d if x in a}
                _d = {}
                for x in d:
                    if x in a:
                        _d[x] = d[x][0]
                # print('supported ctrl:', a)
                # print('supported loaded:', _d)
                return a == _d
            else:
                return False
        else:
            return False

    @key_error
    def _is_valid_driver(self):
        if self._is_valid_system():
            _sys = str(self.system[0])
            _driver_ver = SUPPORTED_VERSIONS[_sys]['SYSTEM']['DRIVER']['VER']
            if self['SYSTEM']['DRIVER']['VER'][0] == _driver_ver:
                d = self['SYSTEM']['DRIVER']
                a = SUPPORTED_VERSIONS[_sys]['SYSTEM']['DRIVER']
                # Does not work on python 2.6
                # _d = {x: d[x][0] for x in d if x in a}
                _d = {}
                for x in d:
                    if x in a:
                        _d[x] = d[x][0]
                # print('supported driver:', a)
                # print('supported loaded:', _d)
                return a == _d
            else:
                return False
        else:
            return False

    @property
    @key_error
    def system(self):
        """
        Returns system version.

        :return: str
        """
        return self['SYSTEM']['VER']

    @property
    @key_error
    def ctrl(self):
        """
        Returns controller version.

        :return: str
        """
        return self['SYSTEM']['CONTROLLER']['VER']

    @property
    @key_error
    def ctrl_dsp(self):
        """
        Returns Controller DSP version.

        :return: str
        """
        return self['SYSTEM']['CONTROLLER']['DSP']

    @property
    @key_error
    def ctrl_fpga(self):
        """
        Returns Controller FPGA version.

        :return: str
        """
        return self['SYSTEM']['CONTROLLER']['FPGA']

    @property
    @key_error
    def ctrl_pcb(self):
        """
        Returns Controller FPGA version.

        :return: str
        """
        return self['SYSTEM']['CONTROLLER']['PCB']
    @property
    @key_error
    def ctrl_mcpu0(self):
        """
        Returns MCPU0 version.

        :return: str
        """
        return self['SYSTEM']['CONTROLLER']['MCPU0']

    @property
    @key_error
    def ctrl_mcpu1(self):
        """
        Returns MCPU1 version.

        :return: str
        """
        return self['SYSTEM']['CONTROLLER']['MCPU1']

    @property
    @key_error
    def ctrl_mcpu2(self):
        """
        Returns MCPU2 version.

        :return: str
        """
        return self['SYSTEM']['CONTROLLER']['MCPU2']

    @property
    @key_error
    def driver(self):
        """
        Returns driver version.

        :return: str
        """
        return self['SYSTEM']['DRIVER']['VER']

    @property
    @key_error
    def driver_dsp(self):
        """
        Returns driver DSP version.

        :return: str
        """
        return self['SYSTEM']['DRIVER']['DSP']

    @property
    @key_error
    def driver_fpga(self):
        """
        Returns driver FPGS version.

        :return: str
        """
        return self['SYSTEM']['DRIVER']['FPGA']

    @property
    @key_error
    def driver_pcb(self):
        """
        Returns driver PCB version.

        :return: (float, str)
        """
        return self['SYSTEM']['DRIVER']['PCB']

    @property
    @key_error
    def driver_io(self):
        """
        Returns driver PCB version.

        :return: (float, str)
        """
        return self['SYSTEM']['DRIVER']['IO']



import pytest
import random

from icepap import IcePAPController

from patch_socket import mock_socket


def get_random_pos():
    return random.randrange(-10000, 10000)


def confirm_m1(m1):
    # Status register 0x00205013
    assert m1.state_present is True
    assert m1.state_moving is False
    assert m1.state_ready is False
    assert m1.state_settling is False
    assert m1.state_outofwin is True
    assert m1.state_warning is False
    assert m1.state_alive is True
    assert m1.state_mode_str == 'OPER'
    assert m1.state_mode_code == 0
    assert m1.state_disabled is True
    assert m1.state_disable_code == 1
    assert m1.state_disable_str == 'Axis not active'
    assert m1.state_indexer_code == 0
    assert m1.state_indexer_str == 'Internal indexer'
    assert m1.state_stop_code == 1
    assert m1.state_stop_str == 'Stop'
    assert m1.state_limit_positive is False
    assert m1.state_limit_negative is False
    assert m1.state_inhome is False
    assert m1.state_5vpower is True
    assert m1.state_verserr is False
    assert m1.state_poweron is False
    assert m1.state_info_code == 0

    assert m1.stopcode == 0
    assert m1.vstopcode == 'No abnormal stop condition'
    assert m1.active is True
    assert m1.mode == 'OPER'
    assert m1.alarm == (False, '')
    assert m1.config == 'toto@pc1_2019/06/17_12:51:24'
    assert m1.id == ('0008.028E.EB82', '4960')

    assert m1.meas_vcc == 80.2165
    assert m1.meas_i == 0.00545881
    assert m1.meas_ia == -0.00723386
    assert m1.meas_ib == -0.000653267
    assert m1.meas_ic == 0
    assert m1.meas_r == -6894.35
    assert m1.meas_ra == -3797.74
    assert m1.meas_rb == -3797.74
    with pytest.raises(RuntimeError):
        m1.meas_rc

    pos = get_random_pos()
    m1.pos = pos
    assert m1.pos == pos
    pos = get_random_pos()
    m1.pos_shftenc = pos
    assert m1.pos_shftenc == pos
    pos = get_random_pos()
    m1.pos_tgtenc = pos
    assert m1.pos_tgtenc == pos
    pos = get_random_pos()
    m1.pos_ctrlenc = pos
    assert m1.pos_ctrlenc == pos
    pos = get_random_pos()
    m1.pos_encin = pos
    assert m1.pos_encin == pos
    pos = get_random_pos()
    m1.pos_inpos = pos
    assert m1.pos_inpos == pos
    # TODO implement exception to configuration mode
    pos = get_random_pos()
    m1.pos_absenc = pos
    assert m1.pos_absenc == pos
    pos = get_random_pos()
    m1.pos_motor = pos
    assert m1.pos_motor == pos
    pos = get_random_pos()
    m1.pos_sync = pos
    assert m1.pos_sync == pos

    pos = get_random_pos()
    m1.enc = pos
    assert m1.enc == pos
    pos = get_random_pos()
    m1.enc_shftenc = pos
    assert m1.enc_shftenc == pos
    pos = get_random_pos()
    m1.enc_tgtenc = pos
    assert m1.enc_tgtenc == pos
    pos = get_random_pos()
    m1.enc_ctrlenc = pos
    assert m1.enc_ctrlenc == pos
    pos = get_random_pos()
    m1.enc_encin = pos
    assert m1.enc_encin == pos
    pos = get_random_pos()
    m1.enc_inpos = pos
    assert m1.enc_inpos == pos
    # TODO implement exception to configuration mode
    pos = get_random_pos()
    m1.enc_absenc = pos
    assert m1.enc_absenc == pos
    pos = get_random_pos()
    m1.enc_motor = pos
    assert m1.enc_motor == pos
    pos = get_random_pos()
    m1.enc_sync = pos
    assert m1.enc_sync == pos

    m1.velocity = 200
    assert m1.velocity == 200
    assert m1.velocity_current == 200
    assert m1.velocity_default == 50
    assert m1.velocity_max == 3000
    assert m1.velocity_min == 2

    with pytest.raises(AttributeError):
        m1.velocity_default = 100
    with pytest.raises(AttributeError):
        m1.velocity_min = 100
    with pytest.raises(AttributeError):
        m1.velocity_max = 100

    m1.acctime = 0.1
    assert m1.acctime == 0.1
    assert m1.acctime_default == 0.01
    assert m1.acctime_steps == 30

    assert m1.pcloop is True
    m1.pcloop = False
    assert m1.pcloop is False

    assert m1.indexer == 'INTERNAL'

    m1.infoa = ['LOW']
    assert set(m1.infoa) == {'LOW', 'NORMAL'}
    m1.infob = ['HIGH', 'NORMAL']
    assert set(m1.infob) == {'HIGH', 'NORMAL'}
    m1.infoc = ['LOW']
    assert set(m1.infoc) == {'LOW', 'NORMAL'}
    m1.outpaux = ['HIGH', 'NORMAL']
    assert set(m1.outpaux) == {'HIGH', 'NORMAL'}
    m1.outpos = ['MOTOR']
    assert set(m1.outpos) == {'MOTOR', 'NORMAL'}
    m1.syncpos = ['AXIS']
    assert set(m1.syncpos) == {'AXIS', 'NORMAL'}
    m1.syncaux = ['ENABLED', 'INVERTED']
    assert set(m1.syncaux) == {'ENABLED', 'INVERTED'}


def ice_auto_axes(f):
    """A helper which provides parametrized auto_axes version of icepap"""
    @pytest.mark.parametrize('auto_axes', [True, False],
                             ids=['smart', 'expert'])
    def wrapper(auto_axes):
        with mock_socket():
            pap = IcePAPController('icepap1', auto_axes=auto_axes)
            return f(pap)
    return wrapper


@pytest.mark.parametrize('auto_axes', [False, True],
                         ids=['expert', 'smart'])
def test_create_eth_icepap(auto_axes):
    with mock_socket():
        with pytest.raises(OSError):
            IcePAPController('weirdhost')
        with pytest.raises(OSError):
            IcePAPController('icepap1', 5001)
        ice = IcePAPController('icepap1', 5000)
        assert ice is not None


@ice_auto_axes
def test_connection(pap):
    assert pap.connected

    pap.disconnect()
    assert not pap.connected


@ice_auto_axes
def test_system(pap):
    assert pap.mode == 'OPER'
    m1 = pap[1]
    confirm_m1(m1)
    m1.pos = 55
    assert pap.get_pos(1) == [55]
    assert pap.get_pos([1]) == [55]
    assert pap.get_pos([1, 5]) == [55, -3]
    assert pap.get_pos([5, 1]) == [-3, 55]

    assert pap.get_fpos(1) == [55]
    assert pap.get_fpos([1]) == [55]
    assert pap.get_fpos([1, 5]) == [55, -3]
    assert pap.get_fpos([5, 1]) == [-3, 55]

    assert pap.get_status(1) == [0x00205013]
    assert pap.get_status([1]) == [0x00205013]
    assert pap.get_status([1, 5]) == [0x00205013, 0x00205013]

    assert pap.get_fstatus(1) == [0x00205013]
    assert pap.get_fstatus([1]) == [0x00205013]
    assert pap.get_fstatus([1, 5]) == [0x00205013, 0x00205013]

    assert pap.get_states([1])[0].status_register == 0x00205013
    assert [s.status_register for s in pap.get_states([1, 5])] == \
           [0x00205013, 0x00205013]

    assert pap.get_power(1) == [True]
    assert pap.get_power([1]) == [True]
    assert pap.get_power([1, 151]) == [True, False]
    assert pap.get_power([151, 1]) == [False, True]

    with pytest.raises(ValueError):
        pap.get_pos('th')

    pap.add_alias('th', 1)

    assert pap.get_pos('th') == [55]


@ice_auto_axes
def test_racks(pap):
    assert pap.get_rid(0) == ['0008.0153.F797']
    assert pap.get_rid([0]) == ['0008.0153.F797']

    assert pap.get_rid([0, 15]) == ['0008.0153.F797', '0008.020B.1028']
    assert pap.get_rid([15, 0]) == ['0008.020B.1028', '0008.0153.F797']

    assert pap.get_rtemp(0) == [30.1]
    assert pap.get_rtemp([0]) == [30.1]

    assert pap.get_rtemp([0, 15]) == [30.1, 29.5]
    assert pap.get_rtemp([15, 0]) == [29.5, 30.1]


@ice_auto_axes
def test_axis_not_plugged(pap):
    with pytest.raises(ValueError):
        pap[200]

    # axis 2 is not installed but is valid
    m2 = pap[2]
    with pytest.raises(RuntimeError):
        m2.pos


def test_smart_axis(smart_pap):
    assert 1 in smart_pap
    assert 'th' not in smart_pap
    assert len(smart_pap.axes) == 3
    assert set(smart_pap.axes) == {1, 5, 151}

    drivers = smart_pap.drivers

    assert len(drivers) == 3

    with pytest.raises(ValueError):
        smart_pap[200]

    with pytest.raises(ValueError):
        smart_pap['m1']

    m1 = smart_pap[1]

    assert m1 in drivers

    smart_pap.add_alias('toto', 1)
    assert len(drivers) == 3
    toto = smart_pap['toto']
    assert m1 is toto

    assert m1.addr == 1
    assert m1.name == 'th'
    assert m1.status == 0x00205013
    assert m1.power is True
    confirm_m1(m1)
    m1.pos = 55
    assert smart_pap.get_pos(1) == [55]
    assert smart_pap.get_pos('toto') == [55]

    del smart_pap[1]
    assert 1 not in smart_pap


def test_smart_find_axes(smart_pap):
    assert set(smart_pap.find_axes()) == {1, 5, 151, 152}
    assert set(smart_pap.find_axes(only_alive=True)) == {1, 5, 151}


def test_smart_update_axes(smart_pap):
    m1 = smart_pap[1]
    smart_pap.update_axes()
    assert m1 is smart_pap[1]


def test_expert_axis(expert_pap):
    assert 1 not in expert_pap
    assert 'th' not in expert_pap
    assert len(expert_pap.axes) == 0

    m1 = expert_pap[1]

    assert 1 in expert_pap

    assert len(expert_pap.axes) == 1
    assert set(expert_pap.axes) == {1}

    drivers = expert_pap.drivers

    assert len(expert_pap.drivers) == 1

    with pytest.raises(ValueError):
        expert_pap[200]

    with pytest.raises(ValueError):
        expert_pap['m1']

    expert_pap.add_alias('toto', 1)
    assert len(drivers) == 1
    toto = expert_pap['toto']
    assert m1 is toto

    assert m1.addr == 1
    assert m1.name == 'th'
    assert m1.status == 0x00205013
    assert m1.power is True
    confirm_m1(m1)
    m1.pos = 55

    assert expert_pap.get_pos(1) == [55]
    assert expert_pap.get_pos('toto') == [55]

    del expert_pap[1]
    assert 1 not in expert_pap


def test_expert_find_axes(expert_pap):
    assert len(expert_pap.axes) == 0
    assert set(expert_pap.find_axes()) == {1, 5, 151, 152}
    assert set(expert_pap.find_axes(only_alive=True)) == {1, 5, 151}
    assert len(expert_pap.axes) == 0


def test_expert_update_axes(expert_pap):
    m1 = expert_pap[1]
    expert_pap.update_axes()
    assert 1 in expert_pap
    assert 5 not in expert_pap
    assert m1 is expert_pap[1]



# -----------------------------------------------------------------------------
# This file is part of icepap (https://github.com/ALBA-Synchrotron/pyIcePAP)
#
# Copyright 2008-2017 CELLS / ALBA Synchrotron, Bellaterra, Spain
#
# Distributed under the terms of the GNU General Public License,
# either version 3 of the License, or (at your option) any later version.
# See LICENSE.txt for more info.
#
# You should have received a copy of the GNU General Public License
# along with icepap. If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------
import time
import collections.abc
import contextlib

from .axis import IcePAPAxis
from .utils import State, is_moving, get_ctrl_item, get_item


class Group:

    def __init__(self, motors):
        if isinstance(motors, IcePAPAxis):
            motors = [motors]
        ctrls = set(motor._ctrl for motor in motors)
        assert len(ctrls) == 1, 'motors must be from same controller'
        self._controller = ctrls.pop()
        self._motors = motors
        self._names = None

    @property
    def controller(self):
        return self._controller

    @property
    def motors(self):
        return self._motors

    @property
    def names(self):
        if self._names is None:
            self._names = self.get_names()
        return self._names

    @property
    def axes(self):
        return [motor.axis for motor in self.motors]

    def get_names(self):
        return get_item(self.motors, "name")

    def get_acctime(self):
        return get_ctrl_item(self.controller.get_acctime, self.axes)

    def get_velocity(self):
        return get_ctrl_item(self.controller.get_velocity, self.axes)

    def get_pos(self):
        return get_ctrl_item(self.controller.get_pos, self.axes)

    def get_fpos(self):
        return get_ctrl_item(self.controller.get_fpos, self.axes)

    def get_states(self):
        return get_ctrl_item(self.controller.get_states, self.axes, State(0))

    def is_moving(self):
        return is_moving(self.get_states())

    def get_power(self):
        return get_ctrl_item(self.controller.get_power, self.axes)

    def start_stop(self):
        self._controller.stop(self.axes)

    def start_move(self, positions, **kwargs):
        args = [[mot.addr, pos] for mot, pos in zip(self._motors, positions)]
        self._controller.move(args, **kwargs)

    def start_rmove(self, positions, **kwargs):
        args = [[mot.addr, pos] for mot, pos in zip(self._motors, positions)]
        self._controller.rmove(args, **kwargs)

    def wait_stopped(self, timeout=None, interval=10e-3):
        """Helper loop to wait for group to finish moving"""

        start = time.time()
        while self.is_moving():
            time.sleep(interval)
            if timeout:
                elapsed = time.time() - start
                if elapsed > timeout:
                    return False
        return True


def group(*objs):
    if len(objs) == 1 and isinstance(objs[0], Group):
        return objs[0]
    motors = []
    for obj in objs:
        if isinstance(obj, IcePAPAxis):
            motors.append(obj)
        elif isinstance(obj, collections.abc.Sequence):
            motors.extend(group(*obj).motors)
        else:
            motors.extend(obj.motors)
    return Group(motors)


def gen_move(group, target):
    with ensure_power(group):
        group.start_move(target)
        for event in gen_motion(group):
            yield event


def gen_rmove(group, target):
    with ensure_power(group):
        group.start_rmove(target)
        for event in gen_motion(group):
            yield event


def gen_motion(group):
    while True:
        states, positions = group.get_states(), group.get_pos()
        yield states, positions
        if not is_moving(states):
            break


@contextlib.contextmanager
def ensure_power(obj, on=True):
    """
    Power context manager. Entering context ensures the motor(s) have power
    (or an exception is thrown). Leaving the context leaves motor(s) has we
    found them. The context manager object is an icepap.Group.

    :param obj: a Group, IcePAPAxis or a sequence of IcePAPAxis
    :param on: if True, ensures power on all motors and when leaving the
               context restores power off on the motors  that were powered
               of before. If False the reverse behavior is applied

    Example::

        from icepap import IcePAPController, ensure_power
        ipap = IcePAPController("ipap.acme.com")
        m1, m2 = ipap[1], ipap[2]
        m1.power = False
        m2.power = True
        assert (m1.power, m2.power) == (False, True)
        with ensure_power((m1, m2)) as group:
            assert (m1.power, m2.power) == (True, True)
            group.start_move((1000, 2000))
            group.wait_move()
        assert (m1.power, m2.power) == (False, True)
    """
    g = group(obj)
    ctrl = g.controller
    powers = g.get_power()
    to_power = [addr for addr, power in zip(g.axes, powers) if power != on]
    if to_power:
        ctrl.set_power(to_power, on)
    try:
        yield g
    finally:
        if to_power:
            ctrl.set_power(to_power, not on)